/* $Id: icongen.c,v 1.2 2004/12/31 08:26:09 ecd Exp $
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <glib.h>
#include <gdk/gdk.h>

#include <pixman.h>

static char *progname;

typedef struct {
	double		x;
	double		y;
} point_t;

typedef struct {
	double		y1;	/* y coordinate of top edge */
	double		x11;	/* x coordinate of top left corner */
	double		x21;	/* x coordinate of top right corner */
	double		y2;	/* y coordinate of bottom edge */
	double		x12;	/* x coordinate of bottom left corner */
	double		x22;	/* x coordinate of bottom right corner */
} trapezoid_t;

static void draw_trapezoids(GdkPixbuf *pixbuf, guint32 color, GList *traps);

static double
i2d(int x)
{
	return (double) x;
}

static int
compare_points(const void *ap, const void *bp)
{
	const point_t *a = ap, *b = bp;

	if (a->y < b->y)
		return -1;
	if (a->y > b->y)
		return 1;
	if (a->x < b->x)
		return -1;
	if (a->x > b->x)
		return 1;
	return 0;
}

static double
compute_x(point_t *p1, point_t *p2, double y)
{
	double dy;

	dy = p2->y - p1->y;
	if (dy == 0.0)
		return p1->x;

	return p1->x + (y - p1->y) * (p2->x - p1->x) / dy;
}

static int
add_trap_from_points(GList **traps, double top, double bottom,
		     double top_left_x, double bottom_left_x,
		     double top_right_x, double bottom_right_x)
{
	trapezoid_t *trap;

	if (top == bottom)
		return 0;

	trap = g_malloc(sizeof(trapezoid_t));
	if (NULL == trap) {
		printf("%s:%u: malloc(trapezoid_t) failed\n",
		       __FUNCTION__, __LINE__);
		return -ENOMEM;
	}

	trap->y1 = top;
	trap->y2 = bottom;

	if (top_left_x > top_right_x) {
		trap->x11 = top_right_x;
		trap->x21 = top_left_x;
	} else {
		trap->x11 = top_left_x;
		trap->x21 = top_right_x;
	}

	if (bottom_left_x > bottom_right_x) {
		trap->x12 = bottom_right_x;
		trap->x22 = bottom_left_x;
	} else {
		trap->x12 = bottom_left_x;
		trap->x22 = bottom_right_x;
	}

	*traps = g_list_append(*traps, trap);
	if (NULL == *traps) {
		printf("%s:%u: g_list_append() failed\n",
		       __FUNCTION__, __LINE__);
		return -ENOMEM;
	}

	return 0;
}

static int
tessellate_triangle(double x1, double y1, double x2, double y2,
		    double x3, double y3, GList **traps)
{
	point_t tsort[3];
	double intersect;
	int err;

	tsort[0].x = x1;
	tsort[0].y = y1;
	tsort[1].x = x2;
	tsort[1].y = y2;
	tsort[2].x = x3;
	tsort[2].y = y3;

	qsort(tsort, 3, sizeof(point_t), compare_points);

	intersect = compute_x(&tsort[0], &tsort[2], tsort[1].y);

	err = add_trap_from_points(traps,
				   tsort[0].y, tsort[1].y,
				   tsort[0].x, intersect,
				   tsort[0].x, tsort[1].x);
	if (err < 0)
		return err;
	err = add_trap_from_points(traps,
				   tsort[1].y, tsort[2].y,
				   intersect, tsort[2].x,
				   tsort[1].x, tsort[2].x);
	if (err < 0)
		return err;

	return 0;
}

static int
tessellate_rectangle(double x1, double y1, double x2, double y2,
		     double x3, double y3, double x4, double y4,
		     GList **traps)
{
	point_t tsort[4];
	double isec02, isec13;
	int err;

	tsort[0].x = x1;
	tsort[0].y = y1;
	tsort[1].x = x2;
	tsort[1].y = y2;
	tsort[2].x = x3;
	tsort[2].y = y3;
	tsort[3].x = x4;
	tsort[3].y = y4;

	qsort(tsort, 4, sizeof(point_t), compare_points);

	isec02 = compute_x(&tsort[0], &tsort[2], tsort[1].y);
	isec13 = compute_x(&tsort[1], &tsort[3], tsort[2].y);

	err = add_trap_from_points(traps,
				   tsort[0].y, tsort[1].y,
				   tsort[0].x, isec02,
				   tsort[0].x, tsort[1].x);
	if (err < 0)
		return err;
	err = add_trap_from_points(traps,
				   tsort[1].y, tsort[2].y,
				   isec02, tsort[2].x,
				   tsort[1].x, isec13);
	if (err < 0)
		return err;
	err = add_trap_from_points(traps,
				   tsort[2].y, tsort[3].y,
				   tsort[2].x, tsort[3].x,
				   isec13, tsort[3].x);
	if (err < 0)
		return err;

	return 0;
}

static int
draw_line(GdkPixbuf *pixbuf, guint32 color, double w,
	  double x1, double y1, double x2, double y2)
{
	GList *traps = NULL;
	double alpha, sina, cosa;
	double halfwidth;
	int err;

	halfwidth = w / 2.0;

	alpha = atan2(y2 - y1, x2 - x1);
	sina = sin(alpha);
	cosa = cos(alpha);

	err = tessellate_rectangle(x1 - halfwidth * sina,
				   y1 + halfwidth * cosa,
				   x1 + halfwidth * sina,
				   y1 - halfwidth * cosa,
				   x2 - halfwidth * sina,
				   y2 + halfwidth * cosa,
				   x2 + halfwidth * sina,
				   y2 - halfwidth * cosa,
				   &traps);
	if (err < 0) {
		printf("%s:%u: error %d: %s\n", __FUNCTION__, __LINE__,
		       err, strerror(-err));
		return err;
	}

	draw_trapezoids(pixbuf, color, traps);

	g_list_foreach(traps, (GFunc) free, NULL);
	g_list_free(traps);

	return 0;
}

static int
draw_filled_circle(GdkPixbuf *pixbuf, guint32 color,
		   double x, double y, double r)
{
	GList *traps = NULL;
	double x1, y1, x2, y2;
	double a, inc;
	int err;

	if (r < 7.0)
		inc = M_PI / 8;
	else
		inc = asin(2.5 / r);

	x2 = x + r * cos(0.0);
	y2 = x - r * sin(0.0);
	for (a = inc; a < 2.0 * M_PI; a += inc) {
		x1 = x2;
		y1 = y2;
		x2 = x + r * cos(a);
		y2 = x + r * sin(a);

		err = tessellate_triangle(x, y, x1, y1, x2, y2, &traps);
		if (err < 0) {
			fprintf(stderr, "%s: tessellate_triangle() failed\n",
				progname);
			return err;
		}
	}
	x1 = x2;
	y1 = y2;
	x2 = x + r * cos(2.0 * M_PI);
	y2 = x + r * sin(2.0 * M_PI);

	err = tessellate_triangle(x, y, x1, y1, x2, y2, &traps);
	if (err < 0) {
		fprintf(stderr, "%s: tessellate_triangle() failed\n",
			progname);
		return err;
	}

	draw_trapezoids(pixbuf, color, traps);

	g_list_foreach(traps, (GFunc) free, NULL);
	g_list_free(traps);

	return 0;
}

static int
draw_circle(GdkPixbuf *pixbuf, guint32 color, double w,
	    double x, double y, double r)
{
	GList *traps = NULL;
	double x1, y1, x2, y2, x3, y3, x4, y4;
	double halfwidth;
	double a, inc;
	int err;

	halfwidth = w / 2.0;

	if (r < 7.0)
		inc = M_PI / 8;
	else
		inc = asin(2.5 / r);

	x3 = x + (r + halfwidth) * cos(0.0);
	y3 = x - (r + halfwidth) * sin(0.0);
	x4 = x + (r - halfwidth) * cos(0.0);
	y4 = x - (r - halfwidth) * sin(0.0);
	for (a = inc; a < 2.0 * M_PI; a += inc) {
		x1 = x3;
		y1 = y3;
		x2 = x4;
		y2 = y4;
		x3 = x + (r + halfwidth) * cos(a);
		y3 = x - (r + halfwidth) * sin(a);
		x4 = x + (r - halfwidth) * cos(a);
		y4 = x - (r - halfwidth) * sin(a);

		err = tessellate_rectangle(x1, y1, x2, y2, x3, y3, x4, y4, &traps);
		if (err < 0) {
			fprintf(stderr, "%s: tessellate_rectangle() failed\n",
				progname);
			return err;
		}
	}
	x1 = x3;
	y1 = y3;
	x2 = x4;
	y2 = y4;
	x3 = x + (r + halfwidth) * cos(2.0 * M_PI);
	y3 = x - (r + halfwidth) * sin(2.0 * M_PI);
	x4 = x + (r - halfwidth) * cos(2.0 * M_PI);
	y4 = x - (r - halfwidth) * sin(2.0 * M_PI);

	err = tessellate_rectangle(x1, y1, x2, y2, x3, y3, x4, y4, &traps);
	if (err < 0) {
		fprintf(stderr, "%s: tessellate_rectangle() failed\n",
			progname);
		return err;
	}

	draw_trapezoids(pixbuf, color, traps);

	g_list_foreach(traps, (GFunc) free, NULL);
	g_list_free(traps);

	return 0;
}

static int
generate_icon(int width, int height, const char *filename)
{
	GdkPixbuf *pixbuf;
	double cx, cy, r;
	double lw, a, a1;
	int err;
	int i;

	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, width, height);

	gdk_pixbuf_fill(pixbuf, 0x00000000);

	cx = i2d(width) / 2.0;
	cy = i2d(height) / 2.0;
	if (width < height)
		r = i2d(width) / 2.0 - 1.0;
	else
		r = i2d(height) / 2.0 - 1.0;

	err = draw_filled_circle(pixbuf, 0xffffffff, cx, cy, r);
	if (err)
		return err;

	lw = 1.0;
	if (r < 10.0)
		lw = 0.75;

	err = draw_circle(pixbuf, 0x808080ff, lw, cx, cy, r + 0.25);
	if (err)
		return err;

	err = draw_circle(pixbuf, 0x808080ff, lw, cx, cy, 2.0 * r / 3.0);
	if (err)
		return err;

	if (r >= 10.0) {
		err = draw_circle(pixbuf, 0x808080ff, lw, cx, cy, 1.0 * r / 3.0);
		if (err)
			return err;
	}

	err = draw_line(pixbuf, 0x808080ff, lw, cx, cy - r, cx, cy + r);
	if (err)
		return err;

	err = draw_line(pixbuf, 0x808080ff, lw, cx - r, cy, cx + r, cy);
	if (err)
		return err;

	for (i = 1; i < 12; i++) {
		if ((i % 3) == 0)
			continue;

		a = i2d(i) * M_PI / 6.0;
		err = draw_line(pixbuf, 0x808080ff, lw,
				cx + r / 6.0 * sin(a),
				cy - r / 6.0 * cos(a),
				cx + r * sin(a), cy - r * cos(a));
		if (err)
			return err;
	}

/*
 * Heading Line:
 */
	err = draw_line(pixbuf, 0x008000ff, 2.0 * lw,
			cx, cy, cx, cy - r);
	if (err)
		return err;


	a = 50.0 * M_PI / 180.0;
	a1 = 90.0 * M_PI / 180.0;
	err = draw_line(pixbuf, 0x008000ff, 2.0 * lw,
			cx + 5.5 * r / 6.0 * sin(a),
			cy - 5.5 * r / 6.0 * cos(a),
			cx + 4.2 * r / 6.0 * sin(a1),
			cy - 4.2 * r / 6.0 * cos(a1));

	a = 52.0 * M_PI / 180.0;
	a1 = 90.0 * M_PI / 180.0;
	err = draw_line(pixbuf, 0x0000c0ff, 2.0 * lw,
			cx + 2.5 * r / 6.0 * sin(a),
			cy - 2.5 * r / 6.0 * cos(a),
			cx + 4.1 * r / 6.0 * sin(a1),
			cy - 4.1 * r / 6.0 * cos(a1));


	a = 50.0 * M_PI / 180.0;
	a1 = 52.0 * M_PI / 180.0;
	err = draw_line(pixbuf, 0x800000ff, 2.0 * lw,
			cx + 5.5 * r / 6.0 * sin(a),
			cy - 5.5 * r / 6.0 * cos(a),
			cx + 2.5 * r / 6.0 * sin(a1),
			cy - 2.5 * r / 6.0 * cos(a1));

	a = 224.5 * M_PI / 180.0;
	err = draw_line(pixbuf, 0x800000ff, 1.0 * lw,
			cx + 2.5 * r / 6.0 * sin(a1),
			cy - 2.5 * r / 6.0 * cos(a1),
			cx + 6.0 * r / 6.0 * sin(a),
			cy - 6.0 * r / 6.0 * cos(a));


	if (!gdk_pixbuf_save(pixbuf, filename, "png", NULL, NULL)) {
		fprintf(stderr, "%s: gdk_pixbuf_save(%s, png) failed\n",
			progname, filename);
		return 1;
	}

	return 0;
}

int
main(int argc, char **argv)
{
	int width, height;
	char *end, *p;

	progname = strrchr(argv[0], '/');
	if (progname)
		progname++;
	else
		progname = argv[0];

	if (argc != 3) {
		fprintf(stderr, "usage: %s <width>x<height> <filename>\n", progname);
		exit(1);
	}

	width = strtoul(argv[1], &end, 10);
	if (argv[1] == end || *end != 'x') {
		fprintf(stderr, "usage: %s <width>x<height> <filename>\n", progname);
		exit(1);
	}
	p = end + 1;

	height = strtoul(p, &end, 10);
	if (p == end || *end != '\0') {
		fprintf(stderr, "usage: %s <width>x<height> <filename>\n", progname);
		exit(1);
	}

	g_type_init();

	return generate_icon(width, height, argv[2]);
}

/************************************************************************/

static unsigned int
d2fixed(double x)
{
	return ((unsigned int) (x * 65556.0));
}

static void
check_bounds(double min, double max, double *value)
{
	if (*value < min)
		*value = min;
	if (*value > max)
		*value = max;
}

/* Fallback rendering code for anti-aliased trapezoids.
 */
static void
draw_trapezoids(GdkPixbuf *pixbuf, guint32 color, GList *traps)
{
	int n_trapezoids = g_list_length(traps);
	pixman_trapezoid_t pixtrap[n_trapezoids];
	pixman_format_t *format;
	pixman_image_t *dst, *src;
	pixman_color_t pixman_color;
	pixman_bits_t *pixman_bits, *b;
	trapezoid_t *trap;
	GList *tp;
	double min_x, max_x, min_y, max_y;
	int x, y, w, h, width, height;
	guint gdk_stride, pixman_stride;
	guint pixman_bits_stride;
	guchar *gdk_data;
	guchar *p, *q;
	guchar alpha;
	guint t;
	int i, j, k, row;

	if (n_trapezoids == 0)
		return;

	/* compute bounding box */
	w = gdk_pixbuf_get_width(pixbuf);
	h = gdk_pixbuf_get_height(pixbuf);

	trap = traps->data;

	min_x = max_x = trap->x11;
	min_y = max_y = trap->y1;

	tp = traps;
	for (i = 0; i < n_trapezoids; i++) {
		trap = tp->data;

		if (trap->x11 < min_x) min_x = trap->x11;
		if (trap->x21 > max_x) max_x = trap->x21;
		if (trap->x12 < min_x) min_x = trap->x12;
		if (trap->x22 > max_x) max_x = trap->x22;
		if (trap->y1 < min_y) min_y = trap->y1;
		if (trap->y2 > max_y) max_y = trap->y2;

		tp = tp->next;
	}


	check_bounds(0.0, i2d(w), &min_x);
	check_bounds(0.0, i2d(w), &max_x);
	check_bounds(0.0, i2d(h), &min_y);
	check_bounds(0.0, i2d(h), &max_y);


	x = floor (min_x);
	width = ceil (max_x) - x;
	y = floor (min_y);
	height = ceil (max_y) - y;


	if (width == 0 || height == 0)
		return;

	pixman_color.red = (((color >> 24) & 0xff) << 8) |
			   (((color >> 24) & 0xff) << 0);
	pixman_color.green = (((color >> 16) & 0xff) << 8) |
			     (((color >> 16) & 0xff) << 0);
	pixman_color.blue = (((color >> 8) & 0xff) << 8) |
			    (((color >> 8) & 0xff) << 0);

	/* draw the trapezoids into the alpha channel */

	format = pixman_format_create(PIXMAN_FORMAT_NAME_A8);
	if (NULL == format) {
		printf("%s:%u: pixman_format_create() failed\n",
		       __FUNCTION__, __LINE__);
		return;
	}

  	dst = pixman_image_create(format, width, height);
	if (NULL == dst) {
		printf("%s:%u: pixman_image_create(dst) failed\n",
		       __FUNCTION__, __LINE__);
		return;
	}

	pixman_format_destroy(format);

	format = pixman_format_create(PIXMAN_FORMAT_NAME_RGB24);
	if (NULL == format) {
		printf("%s:%u: pixman_format_create() failed\n",
		       __FUNCTION__, __LINE__);
		return;
	}

  	src = pixman_image_create(format, 1, 1);
	if (NULL == src) {
		printf("%s:%u: pixman_image_create(src) failed\n",
		       __FUNCTION__, __LINE__);
		return;
	}
	pixman_format_destroy(format);

	pixman_image_set_repeat(src, 1);
	pixman_fill_rectangle(PIXMAN_OPERATOR_OVER, src, &pixman_color, 0, 0, 1, 1);

	tp = traps;
	for (i = 0; i < n_trapezoids; i++) {
		trap = tp->data;

		pixtrap[i].top = d2fixed(trap->y1 - i2d(y));
		pixtrap[i].bottom = d2fixed(trap->y2 - i2d(y));

		pixtrap[i].left.p1.x = d2fixed(trap->x11 - i2d(x));
		pixtrap[i].left.p1.y = d2fixed(trap->y1 - i2d(y));
		pixtrap[i].left.p2.x = d2fixed(trap->x12 - i2d(x));
		pixtrap[i].left.p2.y = d2fixed(trap->y2 - i2d(y));

		pixtrap[i].right.p1.x = d2fixed(trap->x21 - i2d(x));
		pixtrap[i].right.p1.y = d2fixed(trap->y1 - i2d(y));
		pixtrap[i].right.p2.x = d2fixed(trap->x22 - i2d(x));
		pixtrap[i].right.p2.y = d2fixed(trap->y2 - i2d(y));

		tp = tp->next;
	}

	pixman_composite_trapezoids(PIXMAN_OPERATOR_SRC, src, dst, 0, 0,
				    pixtrap, n_trapezoids);


	gdk_data = gdk_pixbuf_get_pixels(pixbuf);
	gdk_stride = gdk_pixbuf_get_rowstride(pixbuf);
	alpha = (color >> 0) & 0xff;

	gdk_data += y * gdk_stride;

	pixman_bits = pixman_image_get_data(dst);
	pixman_stride = pixman_image_get_stride(dst);
	pixman_bits_stride = pixman_stride >> (IC_SHIFT - 3);

	for (row = 0; row < height; row++) {
		b = pixman_bits;

		for (i = 0; i < pixman_bits_stride; i++, b++) {
			if (0 == *b)
				continue;

			k = i << (IC_SHIFT - 3);

			p = (guchar *) b;
			q = gdk_data + 4 * (x + k);

			for (j = 0; j < 4 && j + k < width; j++, p++, q += 4) {
				if (0 == *p)
					continue;

				t = *p * ((color >> 24) & 0xff) +
						(255 - *p) * q[0] + 0x80;
				q[0] = (t + (t >> 8)) >> 8;
				t = *p * ((color >> 16) & 0xff) +
						(255 - *p) * q[1] + 0x80;
				q[1] = (t + (t >> 8)) >> 8;
				t = *p * ((color >> 8) & 0xff) +
						(255 - *p) * q[2] + 0x80;
				q[2] = (t + (t >> 8)) >> 8;
				t = *p * ((color >> 0) & 0xff) +
						(255 - *p) * q[3] + 0x80;
				q[3] = (t + (t >> 8)) >> 8;
			}
		}

		pixman_bits += pixman_bits_stride;
		gdk_data += gdk_stride;
	}

	pixman_image_destroy(dst);
	pixman_image_destroy(src);
}

/************************************************************************/
